/*
  SFBHWI2C.cpp Low-level interface for accessing I2C communications on the LPC23xx.
  Copyright (C) 2010 Sam Bayless.  All rights reserved.

  Modifications for inclusion in SFB core
  Copyright (C) 2010 The Regents of the University of New Mexico.  All rights reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
  USA
*/
#include "SFBHWI2C.h"

#include <LPC23xx.h>
#include "SFBMacros.h"
#include "irq.h"
#include "clock.h"
#include "SFBSerial.h"

#define I2CON_AA      (1UL << 2)
#define I2CON_SI      (1UL << 3)
#define I2CON_STO     (1UL << 4)
#define I2CON_STA     (1UL << 5)
#define I2CON_I2EN      (1UL << 6)



#define I2CONSET      (  0x00)
#define I2STAT        (  0x04)
#define I2DAT         (  0x08)
#define I2ADR         (  0x0C)
#define I2SCLH        (  0x10)
#define I2SCLL        (  0x14)
#define I2CONCLR      (  0x18)

struct PinDescriptor{
  uv32 & pinsel; // The PINSEL register for SDA and SCL (assumed to be the same)
  u8 sdaPos; // The position in PINSEL to specify the SDA function (SCL position is assumed to be SDA + 2)
  u8 pinSelBits; //the bits (00, 01,10, or 11) to set at the position pinselPos in pinsel (assumed to be the same)
};


struct I2CDescriptor {

  uv32 * base;    // The I2C register base address
  u32 pconMask;              // The PCONP bit to power up the I2C
  uv32 & pclkSel;            // The PCLK_SEL register to specify the CCLK : PCLK ratio
  u8 pclkPos;                 // The position in pclkSel to set the ratio

  const PinDescriptor* pins;

  typedef void (*I2CInterruptHandler)();

  I2CInterruptHandler handler; // The interrupt handling function

  u8 interrupt;              // The interrupt number
  u8 priority;               // The interrupt priority
};

//these static variables are used to bridge between the class and the interrupt.
//They'll be set to pointers to the relevant classes when those classes are created.
static SFBHWI2C* interfaces[2];

void I2C1Handler(){
  interfaces[0]->i2cInterrupt();
}
void I2C2Handler(){
  interfaces[1]->i2cInterrupt();
}

static const PinDescriptor WestPins[2]  = {
  {PINSEL0, 0, 3},   //IC21 on WEST_TX_PIN and WEST_RX_PIN
  {PINSEL1, 6, 3}    //IC21 on SOUTH_D3_PIN and SOUTH_D2_PIN
};

//{PINSEL0, 20, 2}
static const PinDescriptor EastPins[1] = {
  {PINSEL0, 20, 2}
};


/* On the IXM, SDA0 and SCL0 are inaccessible.
 * However, the SDA1 and SCL1, and SDA2 and SCL2, are accessible:
 *
 * SDA1 and SCL1 can be mapped to, respectively:
 * Pins 46 and 47 (WEST_TX_PIN and WEST_RX_PIN)
 * Pins 59 and 58 (SOUTH_D3_PIN and SOUTH_D2_PIN)
 *
 * SDA2 and SCL2 can be mapped to, respectively:
 * Pins 48 and 49 (EAST_TX_PIN and EAST_RX_PIN)
 *
 *
 */
//Note: see chapter 21 of the lpc23xx manual for detailed introduction to i2c on that chip.

static const I2CDescriptor I2C[3] = {      /* const to get the i2c descriptors into flash */
  //Note: We're not including IC20 here, since it is not exposed on the IXM
  /* IC21*/  { (uv32*) I2C1_BASE_ADDR, PCI2C1, PCLKSEL1,6,WestPins,
               I2C1Handler,  I2C1_INT, HIGHEST_PRIORITY },

  /* IC22*/  { (uv32*) I2C2_BASE_ADDR, PCI2C2, PCLKSEL1,  20,EastPins ,
               I2C2Handler, I2C2_INT, MEDIUM_PRIORITY }
};

s8 SFBHWI2C::status(){
  return mode;
}

u32 SFBHWI2C::count(){
  return dataCount;
}

void SFBHWI2C::masterTransmit(){
  //See "Master transmitter mode in the manual
  const uv32 * base = I2C[i2c].base;
  switch(REGISTER_OFFSET(base, I2STAT))
    {
    case 0x00://General purpose error
      REGISTER_OFFSET(base, I2CONSET) = I2CON_STO;
      mode = I2C_MODE_ERROR;
      break;
    case 0x08://Start condition transmitted
    case 0x10://Repeated start condition transmitted
      REGISTER_OFFSET(base, I2DAT) = addr<<1;
      dataCount =0;
      REGISTER_OFFSET(base,I2CONCLR) =I2CON_STA;

      break;
    case 0x18://SLA+W transmitted, ACK received
    case 0x28://DATA transmitted, ACK received
      {
        u32 data;
        if (byteSource) 
          data = byteSource(dataCount);
        else if (dataCount < msgLen) 
          data = msg[dataCount];
        else 
          data = 256;
        if (data < 256) {
          REGISTER_OFFSET(base, I2DAT) = data;
          REGISTER_OFFSET(base, I2CONSET) = I2CON_AA;
          ++dataCount;
        } else {
          if (mode == I2C_MODE_TRANSCEIVE) {
            // Ending xmt side of transceive.  Send repeated start condition.
            REGISTER_OFFSET(base,I2CONSET) = I2CON_STA;
            REGISTER_OFFSET(base,I2CONCLR) = I2CON_STO;
            msg = xcvMsg;              // Move the receive args into
            msgLen = xcvLen;           // the standard position
            mode = I2C_MODE_RECEIVE;   // And go pick up in masterReceive, status 0x10..
          } else {
            // End of xmt.  Send stop condition
            REGISTER_OFFSET(base,I2CONCLR) =  I2CON_STA;
            REGISTER_OFFSET(base,I2CONSET) = I2CON_STO;
            mode = I2C_MODE_DONE;
          }
        }
      }
      break;
    case 0x20://SLA+W transmitted, but no ACK received
    case 0x30://DATA transmitted, NOT ACK received
       mode = I2C_MODE_NO_ACK;
       REGISTER_OFFSET(base,I2CONSET) =I2CON_STO;
       break;
     case 0x38://Arbitration lost in SLA+W or DATA
       REGISTER_OFFSET(base, I2CONSET) = I2CON_STO; //it is also an option here to generate a start condition and enter slave reciever mode (according to the manual)
       mode = I2C_MODE_ARBITRATION_ERROR;
       break;
     default:
       mode = I2C_MODE_ERROR;
       REGISTER_OFFSET(base,I2CONSET) = I2CON_STO;
       break;
     }

   REGISTER_OFFSET(base, I2CONCLR) = I2CON_SI;
 }

 void SFBHWI2C::masterReceive(){
   //See "Master Receiver mode in the manual
   const uv32 * base = I2C[i2c].base;
   switch(REGISTER_OFFSET(base, I2STAT))
     {
     case 0x00://General purpose error
       REGISTER_OFFSET(base,I2CONSET) =I2CON_STO;
       mode = I2C_MODE_ERROR;
       break;
     case 0x08://Start condition transmitted
     case 0x10://Repeated start condition transmitted
       REGISTER_OFFSET(base, I2DAT) = (addr << 1) | 1;
       dataCount = 0;
       REGISTER_OFFSET(base,I2CONCLR) = I2CON_STA;
       break;
     case 0x38://Arbitration lost in SLA+W or DATA
       REGISTER_OFFSET(base,I2CONSET) =I2CON_STO;//it is also an option here to generate a start condition and enter slave receiver mode (according to the manual)
       mode = I2C_MODE_ARBITRATION_ERROR;
       break;
     case 0x40://SLA+R transmitted, ACK received
       REGISTER_OFFSET(base,  I2CONSET) =  I2CON_AA;
       break;
     case 0x48://SLA+R transmitted, but no ACK received
       mode = I2C_MODE_NO_ACK;
       REGISTER_OFFSET(base,I2CONSET) = I2CON_STO;
       break;
     case 0x50://DATA received, ACK has been returned
       {
         bool more = false;
         if (byteSink) more = byteSink(REGISTER_OFFSET(base,I2DAT));
         else if (dataCount < msgLen) {
           msg[dataCount++] = REGISTER_OFFSET(base,I2DAT);
           more = true;
         } 
         if (more) 
           REGISTER_OFFSET(base, I2CONSET) =  I2CON_AA; //return ACK
         else {
           mode = I2C_MODE_DONE;
           REGISTER_OFFSET(base, I2CONSET) = I2CON_STO;
         } 
       }
       break;
     case  0x58://DATA received, NOT ACK has been returned
       // XXXX TODO THIS MUST BE UPDATED FOR BYTESINK!
       if(dataCount<msgLen)
         {
           msg[dataCount++] = REGISTER_OFFSET(base,I2DAT);
           REGISTER_OFFSET(base,I2CONSET) =I2CON_STO;
           mode = I2C_MODE_DONE;
         }else{
         mode = I2C_MODE_DONE; //no more room in buffer for message, stop receiving
         REGISTER_OFFSET(base,I2CONSET) = I2CON_STO;
       }
       break;
     default:
       mode = I2C_MODE_ERROR;
       REGISTER_OFFSET(base,I2CONSET) = I2CON_STO;
       break;
     }
   REGISTER_OFFSET(base,I2CONCLR) = I2CON_SI;
 }

 void  SFBHWI2C::i2cInterrupt(void)
 {

   if(mode==I2C_MODE_RECEIVE) {
     masterReceive();
   } else if(mode >= I2C_MODE_TRANSMIT) {  // includes I2C_MODE_TRANSCEIVE
     masterTransmit();
   } else {
     //we're done, or there was an error.
   }

   // Ackley: This is done by the IRQ wrapper in SFBStartup.S:
   // VICVectAddr = 0xFF;

 }


 //Bit rate (in kHz) should be 100 or 400
 void SFBHWI2C::setBitRate(u32 bitrateKHz)
 {
   API_ASSERT_NONZERO(bitrateKHz);

   u32 curCCLK = getCCLK();
   u32 bitrate = bitrateKHz*1000;

   // Determine slowest sufficient PCLK for this bitrate.  

   // We ask for the PCLK to be a multiple of the bitrate if possible,
   // so that we'll be able to hit the requested rate exactly.  Or else
   // we'll do PCLK == curCCLK and have as small error as we can get.

   u32 pclk;
   u8 pclkCode;
   if (curCCLK/8 >= bitrate && (curCCLK/8 % bitrate == 0)) {
     pclk = curCCLK/8;
     pclkCode = 3;
   } else if (curCCLK/4 >= bitrate && (curCCLK/4 % bitrate == 0)) {
     pclk = curCCLK/4;
     pclkCode = 0;
   } else if (curCCLK/2 >= bitrate && (curCCLK/2 % bitrate == 0)) {
     pclk = curCCLK/2;
     pclkCode = 2;
   } else {
     pclk = curCCLK;
     pclkCode = 1;
   }

   API_ASSERT_LESS_EQUAL(2*bitrate, pclk);

   I2C[i2c].pclkSel &= ~(3<<I2C[i2c].pclkPos);
   I2C[i2c].pclkSel |= (pclkCode<<I2C[i2c].pclkPos);

   u32 rate = pclk/bitrate;
   u32 hiCycles = rate/2;
   u32 loCycles = rate-hiCycles;

   //According to the manual, the I2C bitrate will equal the Frequency of PCLK_I2C/ (I2SCLL + I2SCLH)

   //From the manual: "I2SCLH defines the number of PCLK cycles for the SCL high
   //  time, I2SCLL defines the number of PCLK cycles for the SCL low time"
   //They don't necessarily have to be the same value.

   REGISTER_OFFSET(I2C[i2c].base,I2SCLH) = hiCycles;
   REGISTER_OFFSET(I2C[i2c].base,I2SCLL) = loCycles;

 }

 //Note: Address here is NOT pre-shifted to make space for the r/w bit
 void SFBHWI2C::transmit(u8 address, u32 length, u8 *message)
 {
   if(mode == I2C_MODE_UNINITIALIZED) return;

   mode = I2C_MODE_TRANSMIT;
   dataCount =0;
   addr = address;
   msgLen =length;
   msg = message;
   REGISTER_OFFSET(I2C[i2c].base, I2CONSET)=I2CON_STA;
}

void SFBHWI2C::transceive(u8 address, u32 xmtLength, u8 *xmtData, u32 rcvLength, u8 *rcvData)
{
  if(mode == I2C_MODE_UNINITIALIZED) return;

  mode = I2C_MODE_TRANSCEIVE;
  dataCount = 0;
  addr = address;
  msgLen = xmtLength;
  msg = xmtData;
  xcvLen = rcvLength;
  xcvMsg = rcvData;
  REGISTER_OFFSET(I2C[i2c].base, I2CONSET)=I2CON_STA;
}


void SFBHWI2C::init(u32 pins, u32 bitRate)
{
  u32 status;
  suspendInterrupts(status);

  //First, power down the I2C
  PCONP &= ~I2C[i2c].pconMask ;

  // If we're taking over TX/RX pins, reserve the face so packet
  // serial doesn't try to start up (e.g., after setup() finishes).
  if (pins == I2C1_PINS_WEST_TX_RX || pins == I2C2_PINS_EAST_TX_RX) {  // (optimizer avoids the redundancy)
    if (i2c == I2C_1-1) {
      WestFace.reserve();
    } else {
      EastFace.reserve();
    }
  }

  //Install the handler for this interface
  install_irq(I2C[i2c].interrupt, I2C[i2c].handler,I2C[i2c].priority);

  u8 sdaPos = I2C[i2c].pins[pins].sdaPos;
  u8 sclPos = sdaPos+2;//It happens to be the case for the LPC23xx that sclPos is always sdaPos+2 in pinsel.
  I2C[i2c].pins[pins].pinsel &= ~(3<<sdaPos);//clear the existing value
  I2C[i2c].pins[pins].pinsel &= ~(3<<sclPos);//clear the existing value

  u8 bits =   I2C[i2c].pins[pins].pinSelBits;
  I2C[i2c].pins[pins].pinsel |= (bits << sdaPos);
  I2C[i2c].pins[pins].pinsel |= (bits << sclPos);

  //Power up I2C
  PCONP |= I2C[i2c].pconMask ;

  setBitRate(bitRate);

  // disable and reset interface
  REGISTER_OFFSET(I2C[i2c].base, I2CONCLR)= 0xFF; //This disables I2CON_STA, I2CON_STO, I2CON_SI, I2CON_AA, I2CON_I2EN

  // enable interface
  REGISTER_OFFSET(I2C[i2c].base, I2CONSET)= I2CON_I2EN; //As in example 10.1 in the user manual
  //Ok, we're ready
  mode = I2C_MODE_DONE;

  restoreInterrupts(status);
}

//Note: Address here is NOT pre-shifted to make space for the r/w bit
//function will be called as each incoming byte (u8) is received, with the value of that byte.
//length is the maximum number of bytes to receive from the slave (length == 0xFF for receive any number of bits)
void SFBHWI2C::receive(u8 address,u32 maxLength,u8* buffer)
{
  if(mode == I2C_MODE_UNINITIALIZED)
    {
      return;
    }

  mode = I2C_MODE_RECEIVE;
  dataCount =0;
  addr = address;
  msgLen =maxLength;
  msg = buffer;
  REGISTER_OFFSET(I2C[i2c].base, I2CONSET)=I2CON_STA;
}

SFBHWI2C::SFBHWI2C(u32 interface) :
  byteSource(0),
  byteSink(0),
  dataCount(0),
  msgLen(0),
  msg(0),
  i2c(interface-1),
  mode(I2C_MODE_UNINITIALIZED),
  addr(0)
{
  API_ASSERT_MAX(i2c,I2C_2);
  //Store the pointer so it will be accessible to the interrupt handler
  interfaces[i2c] = this;
}

